#ifndef CLEAN_H_INCLUDED
#define CLEAN_H_INCLUDED

#include "vectors.h"



// Returns magnitude of a vector composed by real and imaginary parts
double magnitude(vector<double> c) {
    return sqrt(pow(c[0],2) + pow(c[1],2));
}


// Builds the CLEAN beam, that will be used to convolve the CLEAN components.
// The beam is modelled by a Gaussian function truncated at +/- 5*b_sigma,
// where b_sigma is the standard deviation determined from the HWHM of the
// amplitude of the peak at 0-frequency in the spectral window function WFN.
//
// The time vector should be symmetrized by subtraction of the mean value
// prior to the calculation of windowFunc by the fourier functions, so the
// beam will be a purely real entity.
vector< vector<double> > gaussanRestoringBeam(vector< vector<double> > windowFunc) {
    // Find the Full-width at half maximum (http://en.wikipedia.org/wiki/Full_width_at_half_maximum)
    // Assume that the maximum is located at the 0th element.
    double hmax = 0.5 * magnitude(windowFunc[0]);

    int index = -1; // Init
    double mag = 0; // Init
    // Search for the first element that is equal or lesser than hmax
    for(int i=0; i<windowFunc.size(); i++){
        mag = magnitude(windowFunc[i]);
        if (mag <= hmax) {
            index = i;
            break;
        }
    }

    if (index == -1) {
        throw("Problems searching for the Full-width at half maximum");
    }

    // The element found is lesser (not exactly equal), then make an interpolation to get a better estimation
    double hwidth = 0;
    if (mag < hmax) {
        int sindex = index-1;
        double smag = magnitude(windowFunc[sindex]);
        // Interpolation:
        double slope = 1./(mag - smag);
        // FWHM of WFN
        hwidth = (float)sindex + slope*(hmax-smag);
    } else {
        // The element is exactly the half maximum
        hwidth = index;
    }

    // Fill the restoring beam with a Gaussian characterized by FWHM = hwidth
    // Take into account that the time vector should has been symmetrized (tvec = tvec - tmean) prior to the
    // calculation of WFN by DFOURT, so the beam will be a purely real function.
    double b_sigma = hwidth / sqrt(2.*log(2.));    // beam sigma in index units
    double constant = 1. / (2.*b_sigma*b_sigma);        // Gaussian normalization constant
    int n_beam = (int)(5.*b_sigma) + 1;              // size of restoring beam

    vector< vector<double> > beam;

    // Fill beam with width 2*n_beam-1
    int fulln_beam = 2*n_beam -1;
    for(int i=1; i <= fulln_beam; i++) {

        /*
            double dF = frequencyInc;
            double dphase = TWOPI * dF * (TSUB-TMEAN);
            double gauss = exp(-1 * constant * pow(i,2));
            double phase = dphase * i;
            complex<double> beam(cos(phase), sin(phase));
            beam = gauss * beam;
            // TSUB-TMEAN = 0 because array has been centered, so imaginary part is always 0
        */

        vector<double> e;
        // Half increasing, half decreasing
        if (i <= n_beam) { // increasing
            e.push_back(exp(-1 * constant * pow(i, 2))); // Real
            e.push_back(0); // Imaginary
        } else { // decreasing
            e.push_back(exp(-1 * constant * pow(2*n_beam - i, 2))); // Real
            e.push_back(0); // Imaginary
        }

        beam.push_back(e);
    }

    return beam;
}


// Return the clean component to be used
complex<double> estimateCleanComponent(vector< vector<double> > dirtyDFT, vector< vector<double> > windowFunc, double index) {
    complex<double> alpha;   // Result
    double err = 1.0e-4;     // allowed error in WNORM

    // Find the (index, index) components which produce DFT(index) through WFN
    complex<double> windowFunc_2l(windowFunc[2*index][0], windowFunc[2*index][1]);
    complex<double> dirtyDFT_l(dirtyDFT[index][0], dirtyDFT[index][1]);
    double mag = magnitude(windowFunc[2*index]);
    double wnorm = 1.0 - pow(mag,2);

    // Trap to avoid singularities
    if (wnorm < err) {
        alpha = 0.5 * dirtyDFT_l;
    } else {
        // Equation (25) 'Time Series Analysis with Clean' paper
        alpha = (dirtyDFT_l - windowFunc_2l * conj(dirtyDFT_l)) / wnorm;
    }

    return alpha;

}


// Builds a new vector from the Window:
//
// · Window function: [-3|-2|-1| 0 |+1|+2|+3]
//                        ||           ||
//                        -l           +l
//
// Considering that our windowFunc only has positive frequencies, computeMinusEffect builds a
// new vector from '-l' to '-l + dirtyFunction size'. Cases:
//
//  · Negative positions: As we only have positive freq, take the conjugate of the positive position
//                        considering that the window function is symetric.
//  · Positive positions: As is.
//  · Out of range:       Zero, but it should not happen because window function shoul be 2 times larger
//                        than the dirtyDFT and abs(l) will always be lesser than dirtyDFT size
vector< vector<double> > computeMinusEffect(vector< vector<double> > windowFunc, int ndirt, int l) {
    if (windowFunc.size() != 2*ndirt) throw("Window must be 2 times dirtyDFT.");

    l = -1 * abs(l);
    vector< vector<double> > effect;
    initComplexVector(effect, ndirt); // Fill with 0 and size equal to dirtyDFT

    for(int i=0; i<ndirt; i++) {
        if (abs(l + i) < windowFunc.size()) {
            if (l + i >= 0) {  // Positive zone
                int pos = abs(l + i);
                effect[i][0] = windowFunc[pos][0];
                effect[i][1] = windowFunc[pos][1];
            } else { // Negative zone (conjugate, symetric in comparation with the positive zone that we directly have)
                int pos = (windowFunc.size() - 1) - abs(l + i);
                //complex<double> cc(windowFunc[abs(i-l)][0], windowFunc[abs(i-l)][1]);
                complex<double> cc(windowFunc[pos][0], windowFunc[pos][1]);
                cc = conj(cc);
                effect[i][0] = real(cc);
                effect[i][1] = imag(cc);
            }
        } else {
            effect[i][0] = 0.0;
            effect[i][1] = 0.0;
            throw("A. Should not happen");
        }
    }

    return effect;
}


// Builds a new vector from the Window:
//
// · Window function: [-3|-2|-1| 0 |+1|+2|+3]
//                        ||           ||
//                        -l           +l
//
// Considering that our windowFunc only has positive frequencies, computeMinusEffect builds a
// new vector from '+l' to '+l + dirtyFunction size'. Cases:
//
//  · Positive positions: As is.
//  · Out of range:       Zero, but it should not happen because window function shoul be 2 times larger
//                        than the dirtyDFT and abs(l) will always be lesser than dirtyDFT size.
vector< vector<double> > computePlusEffect(vector< vector<double> > windowFunc, int ndirt, int l) {
    if (windowFunc.size() != 2*ndirt) throw("Window must be 2 times dirtyDFT.");

    l = abs(l);
    vector< vector<double> > effect;
    initComplexVector(effect, ndirt); // Fill with 0 and size equal to dirtyDFT

    for(int i=0; i<ndirt; i++) {
        if ( i+l < windowFunc.size()) {
            int pos = l + i;
            effect[i][0] = windowFunc[pos][0];
            effect[i][1] = windowFunc[pos][1];
        } else {
            effect[i][0] = 0.0;
            effect[i][1] = 0.0;
            throw("B. Should not happen");
        }
    }

    return effect;
}


// Modifies dirtyDFT removing the influence of the Clean Component
void removeInfluenceOfCleanComponent(vector< vector<double> > &dirtyDFT, vector< vector<double> > windowFunc, complex<double> cc, int l) {
    complex<double> czero(0, 0);
    int nwind = windowFunc.size();
    int ndirt = dirtyDFT.size();

    // To save computation time, calculate previously the effect of +l component
    vector< vector<double> > cplus = computePlusEffect(windowFunc, ndirt, l);

    // To save computation time, calculate previously the effect of -l component
    vector< vector<double> > cminus = computeMinusEffect(windowFunc, ndirt, l);

    // Return the realigned, rescaled window function.
    for(int i = 0; i < dirtyDFT.size(); i++) {
        complex<double> cd(dirtyDFT[i][0], dirtyDFT[i][1]);
        complex<double> cp(cplus[i][0], cplus[i][1]);
        complex<double> cm(cminus[i][0], cminus[i][1]);

        // Equation (23) 'Time Series Analysis with Clean' paper
        // dirtyDFT(i) = dirtyDFT(i) - cc * windowFunc(l + i)- conjugate(cc)*windowFunc(-l + i)
        complex<double> new_cd = cd - cc*cp - conj(cc)*cm;

        dirtyDFT[i][0] = real(new_cd);
        dirtyDFT[i][1] = imag(new_cd);
    }

}

// Convolution with center=0 (convolution in the frequency domain is done by multiplication)
// http://star.pst.qub.ac.uk/idl/CONVOL.html
vector< vector<double> > convol(vector< vector<double> > A, vector< vector<double> > K) {
    vector< vector<double> > R;

    // Center = 0
    for(int i=0; i < A.size() ; i++) {
        complex<double> r(0, 0);
        if (i >= K.size() - 1) {
            for (int j=0; j < K.size(); j++) {
                complex<double> a(A[i-j][0], A[i-j][1]);
                complex<double> k(K[j][0], K[j][1]);
                r += a*k;
            }
        } // Otherwise (0,0)

        vector<double> e;
        e.push_back(real(r));
        e.push_back(imag(r));
        R.push_back(e);
    }

    return R;
}

// Add two complex vectors
vector< vector<double> >  addVectors(vector< vector<double> > a, vector< vector<double> > b) {
    vector< vector<double> > result;

    if (a.size() != b.size()) throw("addVectors: Vectors of different length");

    for(int i=0; i<a.size(); i++) {
        complex<double> aterm(a[i][0], a[i][1]);
        complex<double> bterm(b[i][0], b[i][1]);
        complex<double> sum(aterm+bterm);
        vector<double> e;
        e.push_back(real(sum)); // Real
        e.push_back(imag(sum)); // Imaginary
        result.push_back(e);
    }

    return result;
}


// Applies the CLEAN algorithm to the dirtyDFT and returns the cleanDFT
// and the position of the maximum
int clean(vector<double> freq, vector< vector<double> > dirtyDFT, vector< vector<double> > windowFunc, double gain, int nIterations, vector< vector<double> > &cleanDFT) {
    vector< vector<double> > residualDFT = duplicateVector(dirtyDFT);
    vector< vector<double> > cleanComponents;
    initComplexVector(cleanComponents, dirtyDFT.size()); // Fill with 0 and size equal to dirtyDFT

    int max_index = 0;
    double max = 0;
    for(int i = 0; i < nIterations; i++) {
        //cout << "Iteration: " << i << endl;
        // Find current maximum
        max_index = 0;
        max = magnitude(residualDFT[0]);
        for(int j=1; j < residualDFT.size(); j++) {
            double new_max = magnitude(residualDFT[j]);
            if (new_max > max) {
                max_index = j;
                max = new_max;
            }
        }

        // Estimate CLEAN component
        complex<double> cc = gain * estimateCleanComponent(residualDFT, windowFunc, max_index);

        removeInfluenceOfCleanComponent(residualDFT, windowFunc, cc, max_index);

        // Sum and save clean component
        complex<double> actual_cc(cleanComponents[max_index][0], cleanComponents[max_index][1]);
        complex<double> new_cc(actual_cc + cc);
        cleanComponents[max_index][0] += real(new_cc);
        cleanComponents[max_index][1] += imag(new_cc);
    }

    // Construct a clean beam by fitting the Window witha a Gaussian function
    vector< vector<double> > beam = gaussanRestoringBeam(windowFunc);

    // Convolve CCOMP with the BEAM to produce the CLEAN FT.
    cleanDFT = convol(cleanComponents, beam);

    //  Add the final residual vector to preserve noise distribution
    cleanDFT = addVectors(cleanDFT, residualDFT);

    return max_index;
}

#endif // CLEAN_H_INCLUDED
