#ifndef FILES_H_INCLUDED
#define FILES_H_INCLUDED

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <iomanip>

#include "support.h"

// Read data file separated by commas
vector< vector<double> > readFile() {
    // Read file
    vector< vector<double> > measurements;
    ifstream inData ("input/lmcx4.dat");
    string line;
    if (inData.is_open()) {
        while (! inData.eof() ) {
            getline (inData,line);
            // Split columns
            vector<string> measurement = split(line, ',');
            measurement = clean(measurement);

            vector<double> dmeasurement;
            for(int i=0; i<measurement.size(); i++) {
                dmeasurement.push_back(toFloat(measurement[i]));
            }

            if (dmeasurement.size() > 0 ) {
                measurements.push_back(dmeasurement);
            }
        }
    }
    return measurements;
}

// Save plain and simple Fourier transform results with a given x axis in the frequency domain
void saveData(string name, vector<double> freqAxis, vector< vector<double> > data, double sampling_period, double sampling_time, double max_index=-1) {
    string filename = "output/" + name + ".dat";

    ofstream outData;
    outData.open (filename.c_str());
    outData.setf(ios::fixed,ios::floatfield);
    outData.precision(12);

    int N = data.size();
    double sampling_rate = 1 / sampling_period;
    double freq_max = 1 / (2 * sampling_period);
    double fNyquist = freq_max;
    double freq_min = 1/sampling_time;

    outData << "Sampling period = " << sampling_period << endl;
    outData << "Sampling rate = " << sampling_rate << endl;
    outData << "Sampling time = " << sampling_time << endl;
    outData << "Max frequency = " << freq_max << endl;
    outData << "Min frequency = " << freq_min << endl;
    outData << endl;
    if (max_index != -1) {
        outData << "Frequency of more contribution = " << freqAxis[max_index] << "Hz" << endl;
        outData << endl;
    }

    outData << "freq\tmagnitude\ti\tfreq_radians\tfreq_fraction\treal\timg\tphase\tmod^2" << endl;
    // From 0 to N
    for(int i=0; i < N; i++) {
        double freq_fraction = (double)i / N;
        double freq = freqAxis[i];
        double freq_radians = 2*3.14159*freq;
        double real = data[i][0];
        double img = data[i][1];
        double mod_square = (pow(real,2) + pow(img,2));
        double magnitude = sqrt(mod_square);
        double phase = atan2(img,real);

        outData << freq << "\t" << magnitude << "\t";
        outData << i << "\t" << freq_radians << "\t" << freq_fraction << "\t" << real << "\t" << img << "\t";
        outData << phase << "\t" << mod_square << "\t" << endl;
    }
    outData.close();
}

// Save FFTW3 results
void saveData(string name, vector<double> data, int N, double sampling_period, double sampling_time) {
    string filename = "output/" + name + ".dat";

    ofstream outData;
    outData.open (filename.c_str());
    outData.setf(ios::fixed,ios::floatfield);
    outData.precision(12);

    double sampling_rate = 1 / sampling_period;
    double freq_max = 1 / (2 * sampling_period);
    double fNyquist = freq_max;
    double freq_min = 1/sampling_time;

    outData << "Sampling period = " << sampling_period << endl;
    outData << "Sampling rate = " << sampling_rate << endl;
    outData << "Sampling time = " << sampling_time << endl;
    outData << "Max frequency = " << freq_max << endl;
    outData << "Min frequency = " << freq_min << endl;

    outData << endl;
    outData << "freq\tmagnitude\ti\tk\tfreq_radians\tfreq_fraction\treal\timg\tphase\tmod^2" << endl;
    // From -N/2 to N/2, that is why 'k' variable is needed
    for(int i=0; i < N; i++) {
        int k = (-N/2) + i;
        double freq_fraction = (double)k / N;
        double freq = (double)k / (N * sampling_period);
        double freq_radians = 2*3.14159*freq;
        double real = data[2*i];
        double img = data[2*i+1];
        double mod_square = (pow(real,2) + pow(img,2));
        double magnitude = sqrt(mod_square);
        double phase = atan2(img,real);

        // It is symetric for negative values, don't save them
        if (k>=0) {
            outData << freq << "\t" << magnitude << "\t";
            outData << i << "\t" << k << "\t" << freq_radians << "\t" << freq_fraction << "\t" << real << "\t" << img << "\t";
            outData << phase << "\t" << mod_square << "\t" << endl;
        }
    }
    outData.close();
}


// Save NR3 vector and results from Lomb periodogram
void saveData(string name, VecDoub px, VecDoub py, int nout, int jmax, int prob, double sampling_period, double sampling_time) {
    string filename = "output/" + name + ".dat";

    ofstream outData;
    outData.open (filename.c_str());

    outData.setf(ios::fixed,ios::floatfield);
    outData.precision(12);

    double sampling_rate = 1 / sampling_period;
    double freq_max = 1 / (2 * sampling_period);
    double fNyquist = freq_max;
    double freq_min = 1/sampling_time;

    outData << "Sampling period = " << sampling_period << endl;
    outData << "Sampling rate = " << sampling_rate << endl;
    outData << "Sampling time = " << sampling_time << endl;
    outData << "Max frequency = " << freq_max << endl;
    outData << "Min frequency = " << freq_min << endl;
    outData << endl;
    outData << "nout = " << nout << endl;
    outData << "jmax = " << jmax << endl;
    outData << "Frequency of more contribution = px[jmax] = " << px[jmax] << "Hz" << endl;
    outData << "prob = " << prob << " (A small value indicates that a significant periodic signal is present)" << endl;
    outData << endl;

    outData << "freq\tperiodogram" << endl;

    for(int i=0; i < nout; i++) {
        outData << px[i] << "\t" << py[i] << endl;
    }
    outData.close();
}

void saveData(string name, vector< vector<double> > measurementsInPhase) {
    string filename = "output/" + name + ".dat";

    ofstream outData;
    outData.open (filename.c_str());

    outData.setf(ios::fixed,ios::floatfield);
    outData.precision(12);
    outData << "time\tvalue" << endl;

    for(int i=0; i < measurementsInPhase.size(); i++) {
        outData << measurementsInPhase[i][0] << "\t" << measurementsInPhase[i][1] << endl;
    }
    outData.close();
}

#endif // FILES_H_INCLUDED
