#ifndef FOURIER_H_INCLUDED
#define FOURIER_H_INCLUDED

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>
#include <iomanip>
#include <math.h>

#include "vectors.h"

#define TWOPI 6.2831853071796

// Calculates mean and variance for time and data
void dataAnalysis(vector< vector<double> > measurements, double &tmean, double &tvar, double &dmean, double &dvar) {
    double tsum = 0;
    double dsum = 0;

    // Mean
    for(int i=0; i < measurements.size(); i++) {
        tsum += measurements[i][0];
        dsum += measurements[i][1];
    }
    tmean = tsum / measurements.size();
    dmean = dsum / measurements.size();


    double tvsum = 0;
    double dvsum = 0;
    // Variance about the mean of the vector
    for(int i=0; i < measurements.size(); i++) {
        tvsum += pow(measurements[i][0] - tmean, 2);
        dvsum += pow(measurements[i][1] - dmean, 2);
    }
    tvar = tvsum / (measurements.size() - 1);
    dvar = dvsum / (measurements.size() - 1);
}

// Center time and data values using their means
void centerData(vector< vector<double> > measurements, double tmean, double dmean) {
    for(int i=0; i < measurements.size(); i++) {
        measurements[i][0] = measurements[i][0] - tmean;
        measurements[i][1] = measurements[i][1] - dmean;
    }
}

// Determines the frequency parameters used to build the X axis in the frequency domain
void calculateFrequencyParam(vector< vector<double> > measurements, double &freqResolution, double &maxFreq, double &pointPerBeam, int &nfreq, double &freqIncrement) {

    // Check
    for(int i=0; i < measurements.size()-1; i++) {
        if (measurements[i][0] > measurements[i+1][0]) {
            throw("Time vector is not monotonically increasing")
        }
    }

    int last = measurements.size() - 1;
    double sampling_time = (measurements[last][0] - measurements[0][0]);
    freqResolution = 1.0 / (sampling_time); // Minimum frequency

    double minSeparation = measurements[1][0] - measurements[0][0];
    for(int i=1; i < measurements.size(); i++) {
        if (measurements[i][0] - measurements[i-1][0] < minSeparation) {
            minSeparation = measurements[i][0] - measurements[i-1][0];
        }
    }
    double sampling_period = minSeparation;
    double sampling_rate = 1 / sampling_period;
    maxFreq = 1.0 / (2.0 * minSeparation);

    // pointPerBeam: Oversampling factor (such as ofac)
    pointPerBeam = 4; // Usually 4: Optimus value for an afterwards CLEAN algorithm application
    nfreq = (maxFreq / (freqResolution/pointPerBeam)); // * 4
    freqIncrement = freqResolution/pointPerBeam;
}

// Generate the X axis in the frequency domain
vector<double> generateFrequencyList(double nfreq, double freqIncrement) {
    vector<double> freq(2*nfreq);

    freq[0] = 0;
    for(int i=1; i<2*nfreq; i++) {
        freq[i] = freq[i-1] + freqIncrement;
    }

    return freq;
}

double square(int j, int n) {return 1.;}

double bartlett(int j, int n) {return 1.-abs(2.*j/(n-1.)-1.);}

double welch(int j, int n) {return 1.-SQR(2.*j/(n-1.)-1.);}

double hann(int j, int n) {return 0.5*(1.-cos(j*TWOPI/(n-1.)));}

// Compute plain and simple fourier transform using the measurements
// Output:
//  - Frequency vector (x axis in the freq. domain)
//  - Discrete Fourier Transform mixed up with the window function effects that can lead to a wrong interpretation
//    · A complex number per frequency (it can be represented graphically in 2D calculating the modulus)
//  - Window Fourier Transform (WFT): usefull to determine its effects
void dfourt(vector< vector<double> > measurements, double (*windowFunc)(int, int), vector<double> &freq, vector< vector<double> > &dirtyDFT, vector< vector<double> > &WFT) {
    double tmean, tvar, dmean, dvar;
    double freqResolution, maxFreq, pointPerBeam, freqIncrement;
    int nfreq;

    // Obtain time and data mean a variance
    dataAnalysis(measurements, tmean, tvar, dmean, dvar);

    centerData(measurements, tmean, dmean);
    calculateFrequencyParam(measurements, freqResolution, maxFreq, pointPerBeam, nfreq, freqIncrement);

    // X axis in the frequency domain
    freq = generateFrequencyList(nfreq, freqIncrement);

    // Redimension vectors
    initComplexVector(dirtyDFT, nfreq);
    initComplexVector(WFT, 2*nfreq);
    //initComplexVector(WFT, nfreq);

    // For each frequency, calculate Fourier transform
    for(int i=0; i<nfreq; i++) {
        double dftRealSum = 0;
        double dftImgSum = 0;
        double winRealSum = 0;
        double winImgSum = 0;
        for(int n=0; n<measurements.size(); n ++) {
            double phase = -TWOPI * freq[i] * measurements[n][0];
            winRealSum += windowFunc(n, nfreq) * cos(phase);
            winImgSum += windowFunc(n, nfreq) * sin(phase);
            dftRealSum += windowFunc(n, nfreq) * measurements[n][1] * cos(phase);
            dftImgSum += windowFunc(n, nfreq) * measurements[n][1] * sin(phase);
        }
        WFT[i][0] = winRealSum/measurements.size();
        WFT[i][1] = winImgSum/measurements.size();
        dirtyDFT[i][0] = dftRealSum/measurements.size();
        dirtyDFT[i][1] = dftImgSum/measurements.size();
    }

    // Complete the spectral window function (needed if we want to apply the clean algorithm afterwards)
    for(int i=nfreq; i < 2*nfreq; i++) {
        double winRealSum = 0;
        double winImgSum = 0;
        for(int n=0; n<measurements.size(); n ++) {
            double phase = -TWOPI * freq[i] * measurements[n][0];
            winRealSum += windowFunc(n, nfreq) * cos(phase);
            winImgSum += windowFunc(n, nfreq) * sin(phase);
        }
        WFT[i][0] = winRealSum/measurements.size();
        WFT[i][1] = winImgSum/measurements.size();
    }
}



#endif // FOURIER_H_INCLUDED
