#ifndef LEAST_SQUARES_H_INCLUDED
#define LEAST_SQUARES_H_INCLUDED

#include <string>
#include <sstream>

#include <iostream>
#include <iomanip>
#include <fstream>
using namespace std;

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>


struct data {
    size_t n; // Number of measurements
    double * x; // Phase
    double * y; // Measurement
    double * sigma; // Data
};

// Calculates function values of the model function
int expb_f (const gsl_vector * params, void *data, gsl_vector * f) {
    size_t n = ((struct data *) data)->n;
    double *x = ((struct data *) data)->x;
    double *y = ((struct data *) data)->y;
    double *sigma = ((struct data *) data)->sigma;

    // Parameters
    double H = gsl_vector_get (params, 0);
    double A = gsl_vector_get (params, 1);
    double P = gsl_vector_get (params, 2);
    double phi = gsl_vector_get (params, 3);

    size_t i;

    for (i = 0; i < n; i++)     {
        // Function model
        double t = x[i];
        double arg = ((TWOPI / P)*t) + phi;
        double Yi = H + A * cos(arg);
        gsl_vector_set (f, i, (Yi - y[i]) / sigma[i]);
    }

    return GSL_SUCCESS;
}

// Calculates derivatives function model of the model function
int expb_df (const gsl_vector * params, void *data, gsl_matrix * J) {
    size_t n = ((struct data *) data)->n;
    double *x = ((struct data *) data)->x;
    double *y = ((struct data *) data)->y;
    double *sigma = ((struct data *) data)->sigma;

    // Parameters
    double H = gsl_vector_get (params, 0);
    double A = gsl_vector_get (params, 1);
    double P = gsl_vector_get (params, 2);
    double phi = gsl_vector_get (params, 3);

    size_t i;

    for (i = 0; i < n; i++) {
        // Jacobian matrix : differentiation
        // J(i,J) = d(f(i)) / d(param(j))
        // where f(i) = (Yi - yi)/sigma[i],
        //       Yi = H + A * cos(((TWOPI / P)*t) + phi);
        // and the param(j) are the parameters (H, A, P, phi)
        double t = x[i];
        double s = sigma[i];
        double arg = ((TWOPI / P)*t) + phi;

        gsl_matrix_set (J, i, 0, 1 / s);
        gsl_matrix_set (J, i, 1, cos(arg) / s);
        gsl_matrix_set (J, i, 2, (TWOPI*A*t*sin(arg))/(s * pow(P,2)));
        gsl_matrix_set (J, i, 3, (-A*sin(arg)) / s);

        /*
        Maxima:
        f = ((H+A*cos(((2*pi/P)*t)+phi)) - yi) / sigma;
        diff(((H+A*cos(((2*pi/P)*t)+phi)) - yi) / sigma,H,1);
        diff(((H+A*cos(((2*pi/P)*t)+phi)) - yi) / sigma,A,1);
        diff(((H+A*cos(((2*pi/P)*t)+phi)) - yi) / sigma,P,1);
        diff(((H+A*cos(((2*pi/P)*t)+phi)) - yi) / sigma,phi,1);
        */

    }
    return GSL_SUCCESS;
}


// Calculates function values and derivatives of the model function
int expb_fdf (const gsl_vector * params, void *data, gsl_vector * f, gsl_matrix * J) {
    expb_f (params, data, f);
    expb_df (params, data, J);

    return GSL_SUCCESS;
}

// Print parameters
int print_state (ofstream *log, size_t iter, gsl_multifit_fdfsolver * s) {
    log->setf (ios::fixed, ios::floatfield);	// output in fixed format
    log->precision (8);		// digits in doubles

    int width = 12;		// setw width for output
    *log << "iter " << iter << ": "
        << "  x = " << setw (width) << gsl_vector_get (s->x, 0)
        << setw (width) << gsl_vector_get (s->x, 1)
        << setw (width) << gsl_vector_get (s->x, 2)
        << setw (width) << gsl_vector_get (s->x, 3)
        << "  |f(x)| = " << scientific << gsl_blas_dnrm2 (s->f) << endl;

    return 0;
}

// Get parameter estimated value
inline double fit (int i, gsl_multifit_fdfsolver * s) {
  return gsl_vector_get (s->x, i);
}

// Get parameter errors
inline double err (int i, gsl_matrix * covar) {
  return sqrt (gsl_matrix_get (covar, i, i));
}

gsl_multifit_fdfsolver *least_squares(ofstream *log, vector< vector<double> > measurementsInPhase, double period, double tolerance) {
    const gsl_multifit_fdfsolver_type *T;
    gsl_multifit_fdfsolver *s;
    gsl_multifit_function_fdf f;

    // Calculate data mean
    double ysum = 0;
    for(int i=0; i<measurementsInPhase.size(); i++) {
        ysum += measurementsInPhase[i][1];
    }
    double mean = ysum/measurementsInPhase.size();

    // Parameters
    const size_t num_params = 4;
    double params_init[4] = { mean, 1.0, period, 1.0 }; // Initial values
    gsl_vector_view params = gsl_vector_view_array (params_init, num_params);

    // Data
    const size_t n = measurementsInPhase.size(); // Number of input data
    double x[n], y[n], sigma[n];
    struct data d = { n, x, y, sigma};

    // Function model
    f.f = &expb_f;  // Calculates function values
    f.df = &expb_df;    // Calculates derivative function values
    f.fdf = &expb_fdf;  // Calculates function values and derivatives
    f.n = n; // Number of input data
    f.p = num_params; // Number of parameters
    f.params = &d; // Data

    // Convert data vectors to C arrays
    for (int i = 0; i < n; i++) {
      x[i] = measurementsInPhase[i][0];
      y[i] = measurementsInPhase[i][1];
      sigma[i] = 1; // measurementsInPhase[i][2];
    };

    // Initialize least square solver
    T = gsl_multifit_fdfsolver_lmsder;
    s = gsl_multifit_fdfsolver_alloc (T, n, num_params);
    gsl_multifit_fdfsolver_set (s, &f, &params.vector);

    int status;
    unsigned int iter = 0;
    print_state (log, iter, s);
    do {
        iter++;
        status = gsl_multifit_fdfsolver_iterate (s);

        //*log << "status = " << gsl_strerror (status) << endl;
        print_state (log, iter, s);
        if (status) break;

        status = gsl_multifit_test_delta (s->dx, s->x, tolerance, tolerance);
    } while (status == GSL_CONTINUE && iter < 500);

    return s;
}

gsl_matrix *get_parameters_margins(gsl_multifit_fdfsolver * s, int num_params) {
    gsl_matrix *covar = gsl_matrix_alloc (num_params, num_params);
    gsl_multifit_covar (s->J, 0.0, covar);
    //gsl_matrix_fprintf (stdout, covar, "%g");
    return covar;
}

#endif // LEAST_SQUARES_H_INCLUDED
