#include <iostream>
#include <iomanip>
#include <fstream>
using namespace std;

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>

#include <string>
#include <sstream>
#include <vector>
#include <math.h>

#include <sys/time.h>

#include <fftw3.h>
#include "nr3_period.h"
#include "fourier.h"
#include "clean.h"
#include "least_squares.h"

#include "files.h"


// Convert units of the datetime array from days to seconds
void dayToSeconds(vector< vector<double> > measurements) {
    double base_time = measurements[0][0];
    for(int i=0; i<measurements.size(); i++) {
        measurements[i][0] = (measurements[i][0] - base_time) * 24 * 60 * 60;
    }
}

// Identify the minimum separation between time moments
double getMinSeparation(vector< vector<double> > measurements) {
    double minSeparation = measurements[1][0] - measurements[0][0];
    for(int i=1; i < measurements.size(); i++) {
        if (measurements[i][0] - measurements[i-1][0] < minSeparation) {
            minSeparation = measurements[i][0] - measurements[i-1][0];
        }
    }
    return minSeparation;
}


// Generate the X axis in the time domain for representing measurements in phase
vector< vector<double> > allocateMeasurementsInPhase(vector< vector<double> > measurements, double selected_period) {
    vector< vector<double> > measurementsInPhase;

    for(int i=0; i<measurements.size(); i++) {
        vector<double> m;
        int phase_integer_part = (int)((measurements[i][0] - measurements[0][0]) / selected_period);
        double phase = ((measurements[i][0] - measurements[0][0]) / selected_period) - phase_integer_part;
        m.push_back(phase);  // Phase
        m.push_back(measurements[i][1]); // Measure
        m.push_back(measurements[i][2]); // Error
        measurementsInPhase.push_back(m);
    }

    return measurementsInPhase;
}

// Get the current number of seconds and nanoseconds of CPU Process execution
timespec start_clock(string name) {
    timespec tS_start;
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &tS_start);
    cout << name << endl;
    return tS_start;
}

// Compares a CPU Process time with the current CPU Process time
void stop_clock(timespec tS_start) {
    timespec tS_stop;
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &tS_stop);
    double start_time = (double)tS_start.tv_sec + ((double)tS_start.tv_nsec / 10e9); // seconds + nanoseconds (10^-9 seconds)
    double stop_time = (double)tS_stop.tv_sec + ((double)tS_stop.tv_nsec / 10e9);
    cout << ">> CPU time taken: " << stop_time - start_time << " seconds" << endl;
}


// Make a bigger vector with 'times' the actual x axis, so we can get 'times' phases
vector< vector<double> > enlargeMeasurementsInPhase(vector< vector<double> > measurementsInPhase, double times) {
    vector< vector<double> > enlargedMeasurementsInPhase;

    for(int i=0; i<measurementsInPhase.size(); i++) {
        vector<double> m;
        m.push_back(measurementsInPhase[i][0]);  // Phase
        m.push_back(measurementsInPhase[i][1]); // Measure
        m.push_back(measurementsInPhase[i][2]); // Error
        enlargedMeasurementsInPhase.push_back(m);
    }

    // Make it longer show we have 2 phases
    for(int j = 1; j < times; j++) {
        for(int i=0; i<measurementsInPhase.size(); i++) {
            vector<double> m;

            // Control possible floating-point errors
            //if (1.0 + measurementsInPhase[i][0] <= 2) {
                m.push_back(j + measurementsInPhase[i][0]); // Phase
                m.push_back(measurementsInPhase[i][1]); // Measure
                m.push_back(measurementsInPhase[i][2]); // Error
                enlargedMeasurementsInPhase.push_back(m);
            //}

        }
    }
    return enlargedMeasurementsInPhase;
}

int main() {

    timespec tS;

    /// Read input file
    vector< vector<double> > measurements = readFile();
    dayToSeconds(measurements); // Change the units of the time vector (from days to seconds)
    double minSamplingPeriod = getMinSeparation(measurements); // Not uniform, get the minimal separation
    double samplingTime = measurements[measurements.size()-1][0] - measurements[0][0];


    int N = measurements.size(); // It should be power of two but FFTW3 can handle it
    fftw_complex *fftw_data_in = toFFTW_complex(getDataArray(measurements, N));


    /// FFTW3 library for Fast Fourier Transform algorithm
    tS = start_clock("Fast Fourier Transform algorithm with FFTW3");
    fftw_complex *fftw_data_out = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * N);
    fftw_plan p = fftw_plan_dft_1d(N, fftw_data_in, fftw_data_out, FFTW_BACKWARD, FFTW_ESTIMATE);
    fftw_execute(p);
    stop_clock(tS);

    saveData("fftw3_DFT", toVectorDouble(fftw_data_out, N), N, minSamplingPeriod, samplingTime);



    /// Numerical recipes: Unevenly sampled Lomb periodogram
    VecDoub px(N);
    VecDoub py(N);
    int nout = 0;
    int jmax = 0;
    Doub prob = 0;

    int hifac = 1; // Frequencies: up to hifac times the “average” Nyquist frequency (hifac increases the x axis)
    int ofac = 4; // Oversampling factor
    VecDoub data = toVecDoub(getDataArray(measurements, N));
    VecDoub datatime = toVecDoub(getTimeArray(measurements, N));
    tS = start_clock("Unevenly sampled Lomb periodogram with NR3 code");
    period(datatime, onlyReal(data), ofac, hifac, px, py, nout, jmax, prob);
    stop_clock(tS);
    saveData("nr3_lomb_periodogram", px, py, nout, jmax, prob, minSamplingPeriod, samplingTime);


    /// Plain and simple Fourier Transform
    vector<double> freq;
    vector< vector<double> > dirtyDFT;
    vector< vector<double> > WFT;
    // CLEAN parameters
    // WARNING: The CLEAN algorithm only works properly well with the square window
    //          the rest of execution have been included as a demonstration
    double gain = 0.5; // Gain per iteration
    int nIterations = 8;
    double max_index = -1;
    vector< vector<double> > cleanDFT;

    // Square window
    tS = start_clock("Plain DFT with Square window");
    dfourt(measurements, &square, freq, dirtyDFT, WFT);
    stop_clock(tS);
    saveData("PlainDFT_square_dirtySpec", freq, dirtyDFT, minSamplingPeriod, samplingTime);
    saveData("PlainDFT_square_WFT", freq, WFT, minSamplingPeriod, samplingTime);

    tS = start_clock("Clean algorithm with Square window");
    max_index = clean(freq, dirtyDFT, WFT, gain, nIterations, cleanDFT);
    stop_clock(tS);
    saveData("PlainDFT_square_clean", freq, cleanDFT, minSamplingPeriod, samplingTime, max_index);

    // Bartlett window (triangle)
    tS = start_clock("Plain DFT with Bartlett window");
    dfourt(measurements, &bartlett, freq, dirtyDFT, WFT);
    stop_clock(tS);
    saveData("PlainDFT_bartlett_dirtySpec", freq, dirtyDFT, minSamplingPeriod, samplingTime);
    saveData("PlainDFT_bartlett_WFT", freq, WFT, minSamplingPeriod, samplingTime);

    tS = start_clock("Clean algorithm with Bartlett window");
    max_index = clean(freq, dirtyDFT, WFT, gain, nIterations, cleanDFT);
    stop_clock(tS);
    saveData("PlainDFT_barlett_clean", freq, cleanDFT, minSamplingPeriod, samplingTime, max_index);

    // Welch window (smooth curve)
    tS = start_clock("Plain DFT with Welch window");
    dfourt(measurements, &welch, freq, dirtyDFT, WFT);
    stop_clock(tS);
    saveData("PlainDFT_welch_dirtySpec", freq, dirtyDFT, minSamplingPeriod, samplingTime);
    saveData("PlainDFT_welch_WFT", freq, WFT, minSamplingPeriod, samplingTime);

    tS = start_clock("Clean algorithm with Welch window");
    max_index = clean(freq, dirtyDFT, WFT, gain, nIterations, cleanDFT);
    stop_clock(tS);
    saveData("PlainDFT_welch_clean", freq, cleanDFT, minSamplingPeriod, samplingTime, max_index);

    // Hann window (smooth curve with slow increasement/decreasement at the beginning/end)
    tS = start_clock("Plain DFT with Hann window");
    dfourt(measurements, &hann, freq, dirtyDFT, WFT);
    stop_clock(tS);
    saveData("PlainDFT_hann_dirtySpec", freq, dirtyDFT, minSamplingPeriod, samplingTime);
    saveData("PlainDFT_hann_WFT", freq, WFT, minSamplingPeriod, samplingTime);

    tS = start_clock("Clean algorithm with Hann window");
    max_index = clean(freq, dirtyDFT, WFT, gain, nIterations, cleanDFT);
    stop_clock(tS);
    saveData("PlainDFT_hann_clean", freq, cleanDFT, minSamplingPeriod, samplingTime, max_index);


    /// Phase
    double selected_period = 1 / px[jmax]; // For the data used: px[jmax] = 0.033017625961 Hz
    vector< vector<double> > measurementsInPhase = allocateMeasurementsInPhase(measurements, selected_period);
    saveData("lmcx4_in_phase", enlargeMeasurementsInPhase(measurementsInPhase, 2));

    /// Levenberg-Marquardt nonlinear least squares with GSL libraries (GNU Scientific Library)
    ofstream log;
    log.open ("output/least_squares.txt");

    double tolerance = 1e-4;

    tS = start_clock("Levenberg-Marquardt nonlinear least squares with GSL libraries");
    gsl_multifit_fdfsolver *s = least_squares(&log, measurementsInPhase, selected_period, tolerance);
    stop_clock(tS);

    gsl_matrix *covar = get_parameters_margins(s, 4); // 4 parameters

    int width = 7;		// setw width for output
    log.setf (ios::fixed, ios::floatfield);	// output in fixed format
    log.precision (5);		// digits in doubles
    log << endl;
    log << "H      = " << setw (width) << fit (0, s) << " +/- " << setw (width) << err (0, covar) << endl;
    log << "A      = " << setw (width) << fit (1, s) << " +/- " << setw (width) << err (1, covar) << endl;
    log << "P      = " << setw (width) << fit (2, s) << " +/- " << setw (width) << err (2, covar) << endl;
    log << "Phi      = " << setw (width) << fit (3, s) << " +/- " << setw (width) << err (3, covar) << endl;
    int status = gsl_multifit_test_delta (s->dx, s->x, tolerance, tolerance);
    log << "status = " << gsl_strerror (status) << endl;

    gsl_multifit_fdfsolver_free (s);

    return 0;
}


