#ifndef VECTORS_H_INCLUDED
#define VECTORS_H_INCLUDED

#include <vector>
#include <fftw3.h>
#include "nr3.h"

/// C++ Standard vector
// Build an unique array where even position will contain real numbers and
// odd position imaginary numbers.
vector <double> getDataArray(vector< vector<double> > measurements, int N) {
    vector <double> udata(2*N); // 2 doubles per position (real & imaginary)

    for(int i=0; i < N; i++) {
        udata[2*i] = measurements[i][1]; // Real
        udata[2*i+1] = 0.0;  // Imaginary
    }

    return udata;
}

// Build an array with the time serie
vector <double> getTimeArray(vector< vector<double> > measurements, int N) {
    vector <double> datatime(N); // 2 doubles per position (real & imaginary)

    for(int i=0; i < N; i++) {
        datatime[i] = measurements[i][0];
    }

    return datatime;
}

// Initializes adding N complex numbers (z = 0 + i0)
// If the elements of the input vector will be erased
void initComplexVector(vector< vector<double> > &v, int size) {
    v.erase(v.begin(), v.begin() + v.size());
    for(int i=0; i < size; i++) {
        vector<double> e;
        e.push_back(0);
        e.push_back(0);
        v.push_back(e);
    }
}

// Duplicate
vector< vector<double> > duplicateVector(vector< vector<double> > vIn) {
    vector< vector<double> > vOut;
    for(int i=0; i<vIn.size(); i++) {
        vector<double> e;
        e.push_back(vIn[i][0]);
        e.push_back(vIn[i][1]);
        vOut.push_back(e);
    }
    return vOut;
}

/// NR3: Numerical recipes
// Get a NR3 vector with the real values
VecDoub onlyReal(VecDoub vIn) {
    VecDoub vOut(vIn.size()/2);

    for(int i=0; i < vIn.size()/2; i++) {
        vOut[i] = vIn[2*i]; // Real
    }
    return vOut;
}

// Convert C++ standard vector to VecDoub
VecDoub toVecDoub(vector<double> vIn) {
    VecDoub vOut(vIn.size());
    for(int i=0; i<vIn.size(); i++) {
        vOut[i] = vIn[i];
    }
    return vOut;
}

// Convert NR3 vector to C++ standard vector
vector<double> toVectorDouble(VecDoub vIn) {
    vector<double> vOut(vIn.size());

    for(int i=0; i < vIn.size()/2; i++) {
        vOut[2*i] = vIn[2*i]; // Real
        vOut[2*i+1] = vIn[2*i+1]; // Imaginary
    }
    return vOut;
}

// Separate a vector C++ into its different components with NR3 vectors
void divideVector(vector< vector<double> > measurements, VecDoub &x, VecDoub &y, VecDoub &sigma) {
    for(int i = 0; i<measurements.size(); i++) {
        x[i] = measurements[i][0];
        y[i] = measurements[i][1];
        sigma[i] = 1; // measurements[i][2];
    }
}

/// FFW3 library
// Convert C++ standard vector to FFTW complex vector
fftw_complex* toFFTW_complex(vector<double> vIn) {
    fftw_complex* vOut = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * vIn.size()/2);

    for(int i=0; i < vIn.size()/2; i++) {
        vOut[i][0] = vIn[2*i]; // Real
        vOut[i][1] = vIn[2*i+1]; // Imaginary
    }
    return vOut;
}

// Convert FFTW complex vector to C++ standard vector
vector<double> toVectorDouble(fftw_complex* vIn, int N) {
    vector<double> vOut(2*N);

    for(int i=0; i < N; i++) {
        vOut[2*i] = vIn[i][0]; // Real
        vOut[2*i+1] = vIn[i][1]; // Imaginary
    }
    return vOut;
}

#endif // VECTORS_H_INCLUDED
